import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable


def viz_segmentation(X, savefn='', cmap='viridis'):
    """
    Visualise segmentation of image.

    Parameters
    ----------
    X : array
        Segmented image that needs to be visualized.
    savefn : str
        Filename to save figure to, (def = '').
    cmap : str
        Colormap in visualization, (def = 'viridis')

    Returns
    -------
    None

    """
    # Figure options
    fig = plt.figure()
    ax = fig.add_subplot(111)

    # Plot prediction
    image = plt.imshow(X, cmap=cmap)

    # Set axes to unvisible
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)

    # Add colormap
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    plt.colorbar(image, cax=cax)

    # Check whether to save figure
    if savefn:
        fig.savefig(savefn, bbox_inches='tight')

    else:
        plt.show()


def viz_embedding(X, Y, classes=[1, 2, 3], marker='o', alpha=.3, ax=[],
                  savefn=''):
    """
    Visualise embedded patches.

    X : array
        Patches mapped to MRAI-net's representation; num_patches by 2.
    Y : array
        Tissue labels corresponding to X; num_patches by 1.
    classes : list(int)
        Numerical values of tissue labels (def = [1, 2, 3])
    marker : str
        Point marker in scatter plot (def = 'o').
    alpha : float
        Parameter indicating how transparent each marker should be (def = 0.3).
    ax : int
        Axis handle to plot in (def = []).
    savefn : str
        Filename to save figure to (def = '')

    Returns
    -------
    None

    """
    if not ax:
        # Figure options
        fig, ax = plt.subplots()

    # Colors for tissues
    color = ['r', 'b', 'c']

    # Loop over classes
    for k, classk in enumerate(classes):

        # Plot labeled source samples in their embedded representation
        ax.scatter(X[Y == classk, 0],
                   X[Y == classk, 1],
                   c=color[k],
                   marker=marker,
                   alpha=alpha)

    # Check whether to save figure
    if savefn:
        fig.savefig(savefn, bbox_inches='tight')

    else:
        plt.show()



import numpy as np
import matplotlib.pyplot as plt

from mraicnn import MRAIConvolutionalNeuralNetwork
from util import extract_all_patches
from viz import viz_embedding

'''Load data set'''

# Load source MRI-scan and corresponding segmentation
X = np.load('./demos/data/subject01_GE2D_1.5T.npy')
Y = np.load('./demos/data/subject01_segmentation.npy')

# Load target MRI-scan and corresponding segmentation
Z = np.load('./demos/data/subject02_GE2D_3.0T.npy')
U = np.load('./demos/data/subject02_segmentation.npy')

# Note that U is missing a lot of labels
print('Proportion missing labels = ' + str(np.mean(~np.isnan(U.ravel()))))

'''Set up MRAI network'''

# Initialize and compile a small neural network
N = MRAIConvolutionalNeuralNetwork(patch_size=(31, 31),
                                   num_kernels=[8],
                                   kernel_size=[(3, 3)],
                                   dense_size=[16, 8],
                                   batch_size=128,
                                   num_epochs=4,
                                   num_draw=10,
                                   margin=10)

'''Train the net'''

# Call training procedure on source and target data
N.train(X, Y, Z, U, num_targets=1)

'''Map images to MRAI representation.'''

# Extract all source patches and feed them through network.
PX = extract_all_patches(X[0], patch_size=(31, 31), add_4d=True)
HX = N.feedforward(PX, scan_ID=0)

# Map label image to sparse array format
sY = N.matrix2sparse(Y[0])

# Extract all target patches and feed them through network.
PZ = extract_all_patches(Z[0], patch_size=(31, 31), add_4d=True)
HZ = N.feedforward(PZ, scan_ID=1)

# Map label image to sparse array format
sU = N.matrix2sparse(U[0], remove_nans=False)

# Filter out missing target labels
HZ = HZ[~np.isnan(sU[:, 2]), :]
sU = sU[~np.isnan(sU[:, 2]), :]

'''Visualize results'''

# Create figure
fig, ax = plt.subplots(figsize=(15, 10))

# Call visualizer
viz_embedding(HX, sY[:, 2], marker='o', ax=ax)
viz_embedding(HZ, sU[:, 2], marker='x', ax=ax)



import numpy as np
import scipy as sp
import scipy.ndimage as nd

import keras.models as km
import keras.layers as kl
import keras.regularizers as kr
import keras.utils as ku

import sklearn.model_selection as sm
import sklearn.linear_model as sl
import sklearn.svm as sv
import sklearn.feature_extraction.image as sf


def strip_skull(X, M):
    """
    Strip the skull from image X based on mask M.

    Parameters
    ----------
    X : array
        MRI-scan that includes the skull of the subject.
    M : array
        Mask array indicating the skull with True.

    Returns
    -------
    X : array
        MRI-scan with 0's at the mask indices.

    """
    # Check whether mask is a boolean array
    if M.dtype is not np.dtype(bool):
        raise ValueError('Mask should be a logical array.')

    # Check if X contains multiple subjects
    if len(X.shape) > 2:

        # Check for same number of subjects in mask and data.
        if not (X.shape[0] == M.shape[0]):
            raise ValueError('Shape mismatch.')

        # Strip skull, for each subject
        for i in range(X.shape[0]):
            X[i][M[i]] = 0

    else:
        # Strip skull
        X[M] = 0

    return X


def subject2image(fn, imsize=(256, 256), segmentation=False, 
                  classes=[0, 1, 2, 3], normalization=False):
    """
    Load subject images.

    Parameters
    ----------
    fn : str
        Filename of image.
    imsize : tuple(int, int)
        Dimensions of the image to load.
    segmentation : bool
        Indicating whether this is a label matrix.
    classes : list[int]
        List of numerical values of tissue classes.
    normalization : bool
        Whether to normalize the image.

    Returns
    -------
    X : array
        Loaded images, number of subjects by image height by image width

    """
    # Check whether filename is supplied in list form
    if not isinstance(fn, list):
        fn = [fn]

    # Find number of subjects
    num_subjects = len(fn)

    # Preallocate images
    if segmentation:
        # Initialize array of floats
        X = np.zeros((num_subjects, *imsize), dtype='uint8')

    else:
        # Initialize array of floats
        X = np.zeros((num_subjects, *imsize), dtype='float32')

    # Loop over subjects
    for s, fn_s in enumerate(fn):

        # Read binary
        X_s = np.fromfile(fn_s, count=np.prod(imsize), dtype='uint8')

        # Reshape binary list to image
        X_s = nd.rotate(X_s.reshape(imsize), 90)

        # Check for label image
        if segmentation:

            # Restrict classes of segmentations
            for classk in np.setdiff1d(np.unique(X_s), classes):
                X_s[X_s == classk] = 0

        else:

            # Normalize pixels
            if normalization:
                X_s[X_s < 0] = 0
                X_s[X_s > 255] = 255
                X_s = X_s / 255.

        # Add image to image array
        X[s, :, :] = X_s

    return X


def extract_all_patches(X, patch_size=(3, 3), edge=(0, 0), add_4d=False):
    """
    Extract all patches from image.

    Parameters
    ----------
    X : array
        Image to extract patches from.
    patch_size : tuple(int, int)
        Dimensions of patches to extract.
    edge : tuple(int, int)
        Dimensions of the outer edge to ignore during sampling.

    Returns
    -------
    patches : array
        Patches array with number of samples by patch height by patch width.

    """
    # Check whether X is an image
    if len(X.shape) > 2:
        raise ValueError('X is not an image.')

    # Shape
    h, w = X.shape

    # Remove edges
    X = X[edge[0]:h-edge[0], edge[1]:w-edge[1]]

    # Step length from center in patch
    vstep = int((patch_size[0] - 1) / 2)
    hstep = int((patch_size[1] - 1) / 2)

    # Pad image before patch extraction
    X = np.pad(X, pad_width=((vstep, vstep), (hstep, hstep)), mode='constant')

    # Sample patches at grid
    patches = sf.extract_patches_2d(X, patch_size=patch_size)

    # Check whether to add a fourth dimension
    if add_4d:
        patches = patches.reshape((*patches.shape, 1))

    return patches


def set_classifier(X, y, classifier='lr', c_range=np.logspace(-5, 4, 10),
                   num_folds=2):
    """
    Construct classifier with optimal regularization parameter.

    Parameters
    ----------
    X : array
        Data matrix, number of samples by number of features.
    y : array
        Label vector or matrix, number of samples by 1, or number of samples by
        number of classes.
    classifier : str
        Type of classifier to use, options are: 'lr' (logistic regression),
        'linsvc' (linear support vector classifier),
        'rbfsvc' (radial basis function support vector classifier),
        'cnn' (convolutional neural network), (def='lr').
    c_range : array
        Range of possible regularization parameters,
        (def=np.logspace(-5, 4, 10)).
    num_folds : int
        Number of folds to use in cross-validation.

    Returns
    -------
    sklearn classifier
        A trained classifier with a predict function.

    """
    # Data shape
    N = X.shape[0]

    # Number of classes
    if len(y.shape) == 2:

        # Check number of classes by number of columns of label matrix
        num_classes = y.shape[1]

    else:
        # Check number of classes by the number of unique entries
        num_classes = len(np.unique(y))

    if classifier == 'lr':

        # Grid search over regularization parameters C
        if N > num_classes:

            # Set up a grid search cross-validation procedure
            modelgs = sm.GridSearchCV(sl.LogisticRegression(),
                                      param_grid=dict(C=c_range),
                                      cv=num_folds)

            # Fit grid search model
            modelgs.fit(X, y)

            # Extract best regularization parameter
            bestC = modelgs.best_estimator_.C

        else:
            # Standard regularization parameter
            bestC = 1

        # Set up model with optimal C
        return sl.LogisticRegression(C=bestC)

    elif classifier == 'linsvc':

        # Grid search over regularization parameters C
        if N > num_classes:

            # Set up a grid search cross-validation procedure
            modelgs = sm.GridSearchCV(sv.LinearSVC(),
                                      param_grid=dict(C=c_range),
                                      cv=num_folds)

            # Fit grid search model
            modelgs.fit(X, y)

            # Extract best regularization parameter
            bestC = modelgs.best_estimator_.C

        else:
            # Standard regularization parameter
            bestC = 1

        # Train model with optimal C
        return sv.LinearSVC(C=bestC)

    elif classifier == 'rbfsvc':

        # Grid search over regularization parameters C
        if N > num_classes:

            # Set up a grid search cross-validation procedure
            modelgs = sm.GridSearchCV(sv.SVC(), cv=num_folds,
                                      param_grid=dict(C=c_range))

            # Fit grid search model
            modelgs.fit(X, y)

            # Extract best regularization parameter
            bestC = modelgs.best_estimator_.C

        else:
            # Standard regularization parameter
            bestC = 1

        # Train model with optimal C
        return sv.SVC(C=bestC)

    elif classifier == 'cnn':

        # Start sequential model
        net = km.Sequential()

        # Convolutional part
        net.add(kl.Conv2D(8, kernel_size=(3, 3),
                          activation='relu',
                          padding='valid',
                          kernel_regularizer=kr.l2(0.001),
                          input_shape=(X.shape[1], X.shape[2], 1)))
        net.add(kl.MaxPooling2D(pool_size=(2, 2), padding='valid'))
        net.add(kl.Dropout(0.2))
        net.add(kl.Flatten())

        # Fully-connected part
        net.add(kl.Dense(16, activation='relu',
                         kernel_regularizer=kr.l2(0.001)))
        net.add(kl.Dropout(0.2))
        net.add(kl.Dense(8, activation='relu',
                         kernel_regularizer=kr.l2(0.001)))
        net.add(kl.Dense(num_classes, activation='softmax'))

        # Compile network architecture
        net.compile(loss='categorical_crossentropy',
                    optimizer='rmsprop',
                    metrics=['accuracy'])

        return net

    else:
        print('Classifier unknown')


def classify(X, y, val=[], classifier='lr', c_range=np.logspace(-5, 4, 10),
             num_folds=2, num_epochs=2, batch_size=32, verbose=True):
    """
    Classify sets of patches.

    Parameters
    ----------
    X : array
        Data matrix, number of samples by number of features.
    y : array
        Label vector or matrix, number of samples by 1, or number of samples by
        number of classes.
    val : list[X, y]
        Validation data and label set, (def = []).
    classifier : str
        Type of classifier to use, options are: 'lr' (logistic regression),
        'linsvc' (linear support vector classifier),
        'rbfsvc' (radial basis function support vector classifier),
        'cnn' (convolutional neural network), (def='lr').
    c_range : array
        Range of possible regularization parameters,
        (def=np.logspace(-5, 4, 10)).
    num_folds : int
        Number of folds to use in cross-validation.
    num_epochs : int
        Number of epochs in training a neural network.
    batch_size : int
        Size of the batches to cut the data set into.
    verbose : bool
        Verbosity during training.

    Returns
    -------
    err : float
        Classification error.

    """
    # Number of classes
    num_classes = len(np.unique(y))

    if classifier == 'cnn':

        # Switch labels to categorical array
        y = ku.to_categorical(y - np.min(y), num_classes)

    else:
        # If data has more than 2 dimensions, reshape it back to 2.
        if len(X.shape) > 2:
            X = X.reshape((X.shape[0], -1))

    # Train regularized classifier
    model = set_classifier(X, y, classifier=classifier, num_folds=num_folds,
                           c_range=c_range)

    if classifier.lower() in ['lr', 'linsvc', 'rbfsvc']:

        if val:

            # Number of validation samples
            N = val[0].shape[0]

            # Error on validation data
            err = 1 - model.fit(X, y).score(val[0].reshape((N, -1)), val[1])

        else:
            # Error cross-validated over training data
            err = np.mean(1 - sm.cross_val_score(model, X, y=y, cv=num_folds))

    elif classifier == 'cnn':

        if val:
            # Validation
            valy = ku.to_categorical(val[1] - np.min(val[1]), num_classes)

            # Fit model
            model.fit(X, y, batch_size=batch_size, epochs=num_epochs,
                      validation_split=0.2, shuffle=True)

            # Error on validation data
            err = 1 - model.test_on_batch(val[0], valy)[1]

        else:
            # Error cross-validated over training data
            err = np.mean(1 - sm.cross_val_score(model, X, y=y, cv=num_folds))
    else:
        raise ValueError('Classifier unknown')

    # Report
    if verbose:
        print('Classification error = ' + str(err))

    return err


def CMA_to_4classes(L):
    """
    Map CMA's automatic segmentation to {BCK,CSF,GM,WM}.

    CMA's automatic segmentation protocol lists over 80 different tissue
    classes. Here we map these back to the four we used in the paper:
    background (BCK), cerebrospinal fluid (CSF), gray matter (GM) and
    white matter (WM). Sets Brainstem and Cerebellum to background
    (16=0, 6,7,8,45,46,47=0).

    Parameters
    ----------
    L : array
        Label matrix

    Returns
    -------
    L : array
        Label matrix

    """
    # Number of subjects
    nI = L.shape[0]

    # Re-map to
    L = -L
    for i in range(nI):
        L[i][L[i] == -0] = 0
        L[i][L[i] == -1] = 0
        L[i][L[i] == -2] = 3
        L[i][L[i] == -3] = 2
        L[i][L[i] == -4] = 1
        L[i][L[i] == -5] = 1
        L[i][L[i] == -6] = 0
        L[i][L[i] == -7] = 0
        L[i][L[i] == -8] = 0
        L[i][L[i] == -9] = 2
        L[i][L[i] == -10] = 2
        L[i][L[i] == -11] = 2
        L[i][L[i] == -12] = 2
        L[i][L[i] == -13] = 2
        L[i][L[i] == -14] = 1
        L[i][L[i] == -15] = 1
        L[i][L[i] == -16] = 0
        L[i][L[i] == -17] = 2
        L[i][L[i] == -18] = 2
        L[i][L[i] == -19] = 2
        L[i][L[i] == -20] = 2
        L[i][L[i] == -21] = 0
        L[i][L[i] == -22] = 0
        L[i][L[i] == -23] = 2
        L[i][L[i] == -24] = 1
        L[i][L[i] == -25] = 0
        L[i][L[i] == -26] = 2
        L[i][L[i] == -27] = 2
        L[i][L[i] == -28] = 2
        L[i][L[i] == -29] = 0
        L[i][L[i] == -30] = 0
        L[i][L[i] == -31] = 0
        L[i][L[i] == -32] = 0
        L[i][L[i] == -33] = 0
        L[i][L[i] == -34] = 0
        L[i][L[i] == -35] = 0
        L[i][L[i] == -36] = 0
        L[i][L[i] == -37] = 0
        L[i][L[i] == -38] = 0
        L[i][L[i] == -39] = 0
        L[i][L[i] == -40] = 0
        L[i][L[i] == -41] = 3
        L[i][L[i] == -42] = 2
        L[i][L[i] == -43] = 1
        L[i][L[i] == -44] = 1
        L[i][L[i] == -45] = 0
        L[i][L[i] == -46] = 0
        L[i][L[i] == -47] = 0
        L[i][L[i] == -48] = 2
        L[i][L[i] == -49] = 2
        L[i][L[i] == -50] = 2
        L[i][L[i] == -51] = 2
        L[i][L[i] == -52] = 2
        L[i][L[i] == -53] = 2
        L[i][L[i] == -54] = 2
        L[i][L[i] == -55] = 2
        L[i][L[i] == -56] = 2
        L[i][L[i] == -57] = 0
        L[i][L[i] == -58] = 2
        L[i][L[i] == -59] = 2
        L[i][L[i] == -60] = 2
        L[i][L[i] == -61] = 0
        L[i][L[i] == -62] = 0
        L[i][L[i] == -63] = 0
        L[i][L[i] == -64] = 0
        L[i][L[i] == -65] = 0
        L[i][L[i] == -66] = 0
        L[i][L[i] == -67] = 0
        L[i][L[i] == -68] = 0
        L[i][L[i] == -69] = 0
        L[i][L[i] == -70] = 0
        L[i][L[i] == -71] = 0
        L[i][L[i] == -72] = 0
        L[i][L[i] == -73] = 0
        L[i][L[i] == -74] = 0
        L[i][L[i] == -75] = 0
        L[i][L[i] == -76] = 0
        L[i][L[i] == -77] = 0
        L[i][L[i] == -78] = 0
        L[i][L[i] == -79] = 0
        L[i][L[i] == -80] = 0
        L[i][L[i] == -81] = 0
        L[i][L[i] == -82] = 0
        L[i][L[i] == -83] = 0
        L[i][L[i] == -84] = 3

    return L



#!/usr/bin/env python3

import numpy as np
import nibabel as nib
import tensorflow as tf
from keras import backend as bK
import sklearn as sk
import sklearn.model_selection as skms
import sklearn.linear_model as sklm
from sklearn.feature_extraction.image import extract_patches_2d


class MRAIDenseNetwork():
    """
    Class of neural networks to learn MRI-scanner acquisition-invariant
    representations.
    """
    def __init__(self, margin=1, imwidth=256, imheight=256,
                 classes=np.array([1, 2, 3]), patch_size=(3, 3), num=2,
                 sS=10, l2=0.001, nT=1):
        """
        Initialize the network with data shape, architecture and
        optimization parameters.
        """
        # Data shape parameters
        self.imwwidth = imwidth
        self.imheight = imheight
        self.patch_size = patch_size
        self.patch_step = ((patch_dim[0]-1)/2, (patch_dim[1]-1)/2)
        self.patch_unraveled = (2*pD + 1)**2
        self.nT = nT

        # Classes
        self.classes = classes

        self.nE = nE
        self.sS = sS

        # Network hyperparameters
        self.margin = margin
        self.l2 = l2

        # Optimization parameters
        self.max_iter = max_iter

        # Initialize net architecture
        self.net = []

    def extract_random_patches(self, im, segm):
        """Extract a random set of patches from image."""

        # Shapes
        num_ims = im.shape[0]
        num_classes = len(self.classes)

        # Preallocate input and output patches
        Ip = np.zeros((nI, nK, self.sS, self.pU+1))
        Il = np.zeros((nI, nK, self.sS), dtype='int64')

        for i in range(nI):
            for k in range(nK):

                # Find tissue in image
                Lik = np.argwhere(L[i][self.pD:-self.pD-1,self.pD:-self.pD-1]==self.K[k])+self.pD

                # Subsample indices
                ixk = np.random.choice(np.ravel_multi_index(Lik.T, (self.w,self.h)), size=self.sS, replace=False)

                # Sample patch from class k
                pIk = self.index2patch(I[i], index=ixk).reshape((self.sS,-1))

                # Augment patches with scanner ID and store
                Ip[i,k,:,:] = np.append(pIk,scan_ID*np.ones((self.sS,1)),axis=1)

                # Store class labels
                Il[i,k,:] = self.K[k]*np.ones((self.sS,))

        return Ip,Il

    def extract_all_patches(self, im, scan_ID=0, patch_size=(3, 3),
                            edge=(0, 0)):
        """Extract all patches from image, and append a scanner ID var."""

        # Shapes
        imheight, imwidth = im.shape

        # Cut edges of image
        im = im[edge[0]:-edge[0], edge[1]:-edge[1]]

        # Sample patches at grid
        im_p = extract_patches_2d(im, patch_size=self.patch_size)

        # Augment patches with scanner ID and store
        return np.append(im_p, scan_ID*np.ones((num_patches, 1)), axis=1)


    def index2patch(self, I, index=np.empty((0,))):
        ''' Slice a patch from an image at a particular index '''

        # Number of patches to sample
        num_samples = index.shape[0]

        # Find 2D index of linear index
        w = np.empty((num_samples,), dtype='int64')
        h = np.empty((num_samples,), dtype='int64')
        if len(index.shape)==1:
            ix = np.empty((num_samples,2), dtype='int64')
            for n in range(num_samples):
                w[n],h[n] = np.unravel_index(index[n], (self.w,self.h))
        elif len(index.shape)==2:
            w = index[:,0]
            h = index[:,1]
        else:
            print('index has wrong shape')

        # Slice out patch(es)
        patches = np.empty((num_samples, self.pW, self.pW))
        for n in range(num_samples):
            patches[n, :, :] =  I[w[n]-self.pD:w[n]+self.pD+1, h[n]-self.pD:h[n]+self.pD+1]

        return patches

    def contrastive_loss(self, y_true, y_pred):
        """Contrastive loss; squared for similar, """
        return bK.sum(y_true * bK.square(y_pred) +
                      (1-y_true) * bK.abs(bK.maximum(self.m - y_pred, 0)))

    def l1_norm(self, vects):
        """l1-norm for loss layer."""
        return bK.sum(bK.abs(vects[0]-vects[1]), axis=1, keepdims=True)

    def l2_norm(self, x):
        """l2-norm for loss layer."""
        return bK.sqrt(bK.maximum(bK.sum(bK.square(x[0]-x[1]), axis=1,
                                         keepdims=True), bK.epsilon()))

    def index_comb(self, A):
        """Generate combinations of two index arrays."""
        return np.array(np.meshgrid(A[0], A[1])).T.reshape(-1, 2)

    def set_net(self, hidden_nodes=[64,16,2], act='relu', opt='rmsprop'):
        ''' Define network architecture '''

        # Setup input layers
        A = ks.layers.Input(shape=(self.pU+1,))
        B = ks.layers.Input(shape=(self.pU+1,))

        # Define a sequential model for a pipeline
        pipe = ks.models.Sequential()
        pipe.add(ks.layers.Dense(hidden_nodes[0], input_shape=(self.pU+1,), activation=act))
        for h in range(1,len(hidden_nodes)-1):
            pipe.add(ks.layers.Dropout(0.2))
            pipe.add(ks.layers.Dense(hidden_nodes[h], activation=act, kernel_regularizer=ks.regularizers.l2(self.l2)))
        pipe.add(ks.layers.Dense(hidden_nodes[-1]))

        print(pipe.summary())

        # Distance in embedding space
        distance = ks.layers.Lambda(self.l1_norm, output_shape=(1,))([pipe(A), pipe(B)])

        # Set model
        model = ks.models.Model(inputs=[A,B], outputs=distance)

        # Compile model with optimizer
        model.compile(loss=self.contrastive_loss, optimizer=opt)

        # Return compiled model
        self.net = model


    def train(self, I,L,J,tix):
        ''' Train the net '''

        # Number of subjects
        nI = I.shape[0]
        nJ = J.shape[0]

        # Number of classes
        nK = len(self.K)

        # Minimum of the number of target labels and sample subset
        minTS = min(self.nT, self.sS)

        # Loop over source subjects
        for i in range(nI):
            print('At source subject '+str(i+1)+'/'+str(nI))

            for j in range(nJ):
                print('At target subject '+str(j+1)+'/'+str(nJ))

                # Pre-allocated data
                train_I = np.empty((2*nK**2*self.sS*minTS,self.pU+1))
                train_J = np.empty((2*nK**2*self.sS*minTS,self.pU+1))
                train_Y = np.empty((2*nK**2*self.sS*minTS,1))

                # Find random other source subject
                if nI==1:
                    o = i
                else:
                    o = np.random.choice(np.setdiff1d(np.arange(nI),i), size=1, replace=False)[0]

                # Loop over classes
                c = 0
                for k in range(nK):

                    # Find tissue in image
                    Lik = np.argwhere(L[i][self.pD:-self.pD-1,self.pD:-self.pD-1]==self.K[k])+self.pD
                    Lok = np.argwhere(L[o][self.pD:-self.pD-1,self.pD:-self.pD-1]==self.K[k])+self.pD

                    # Subsample indices of tissues
                    ixk = np.random.choice(np.ravel_multi_index(Lik.T, (self.w,self.h)), size=self.sS, replace=False)
                    ixo = np.random.choice(np.ravel_multi_index(Lok.T, (self.w,self.h)), size=minTS, replace=False)
                    ixt = np.random.choice(tix[j,k,:].reshape((-1,)), size=minTS, replace=False)

                    # Generate pairwise index combinations
                    ixkk = self.index_comb((ixk,ixt))
                    ixko = self.index_comb((ixk,ixo))

                    # Sample patches sets
                    pIkk = self.index2patch(I[i], index=ixkk[:,0]).reshape(ixkk.shape[0],-1)
                    pJkk = self.index2patch(J[j], index=ixkk[:,1]).reshape(ixkk.shape[0],-1)
                    pIok = self.index2patch(I[o], index=ixko[:,1]).reshape(ixkk.shape[0],-1)

                    # Augment patch with scanner ID
                    pIkk = np.append(pIkk,np.zeros((pIkk.shape[0],1)),axis=1)
                    pJkk = np.append(pJkk, np.ones((pJkk.shape[0],1)),axis=1)
                    pIok = np.append(pIok,np.zeros((pIok.shape[0],1)),axis=1)

                    # Store patches source-target kk
                    train_I[c*self.sS*minTS:(c+1)*self.sS*minTS,:] = pIkk
                    train_J[c*self.sS*minTS:(c+1)*self.sS*minTS,:] = pJkk
                    train_Y[c*self.sS*minTS:(c+1)*self.sS*minTS,:] = np.ones((self.sS*minTS,1))
                    c += 1

                    # Store patches source-source kk
                    train_I[c*self.sS*minTS:(c+1)*self.sS*minTS,:] = pIkk
                    train_J[c*self.sS*minTS:(c+1)*self.sS*minTS,:] = pIok
                    train_Y[c*self.sS*minTS:(c+1)*self.sS*minTS,:] = np.ones((self.sS*minTS,1))
                    c += 1

                    for l in np.setdiff1d(np.arange(nK),k):

                        # Find tissue in image
                        Lol = np.argwhere(L[o][self.pD:-self.pD-1,self.pD:-self.pD-1]==self.K[l])+self.pD

                        # Subsample indices of other tissues
                        ixo = np.random.choice(np.ravel_multi_index(Lol.T, (self.w,self.h)), size=minTS, replace=False)
                        ixt = np.random.choice(tix[j,l,:].reshape((-1,)), size=minTS, replace=False)

                        # Generate pairwise index combinations
                        ixkl = self.index_comb((ixk,ixt))
                        ixol = self.index_comb((ixk,ixo))

                        # Sample patches sets
                        pIkl = self.index2patch(I[i], index=ixkl[:,0]).reshape(ixkl.shape[0],-1)
                        pJkl = self.index2patch(J[j], index=ixkl[:,1]).reshape(ixkl.shape[0],-1)
                        pIol = self.index2patch(I[o], index=ixol[:,1]).reshape(ixol.shape[0],-1)

                        # Augment patch with scanner ID
                        pIkl = np.append(pIkl,np.zeros((pIkl.shape[0],1)),axis=1)
                        pJkl = np.append(pJkl, np.ones((pJkl.shape[0],1)),axis=1)
                        pIol = np.append(pIol,np.zeros((pIol.shape[0],1)),axis=1)

                        # Store sampled patches
                        train_I[c*self.sS*minTS:(c+1)*self.sS*minTS,:] = pIkl
                        train_J[c*self.sS*minTS:(c+1)*self.sS*minTS,:] = pJkl
                        train_Y[c*self.sS*minTS:(c+1)*self.sS*minTS,:] = np.zeros((self.sS*minTS,1))
                        c += 1

                        # Store sampled patches
                        train_I[c*self.sS*minTS:(c+1)*self.sS*minTS,:] = pIkl
                        train_J[c*self.sS*minTS:(c+1)*self.sS*minTS,:] = pIol
                        train_Y[c*self.sS*minTS:(c+1)*self.sS*minTS,:] = np.zeros((self.sS*minTS,1))
                        c += 1

                # Report loss
                self.net.fit(x=[train_I,train_J], y=train_Y, epochs=self.nE, shuffle=True)


    def propagate(self, Ip):
        ''' Propagate a set of patches through acqinv-net '''

        # Define intermediate-output model
        inter_model = ks.models.Model(inputs=self.net.input, outputs=self.net.get_layer(index=2).get_output_at(1))

        # Feed forward
        return inter_model.predict([Ip, Ip])


    def segmentImage(self, model, I, clfr, scan_ID=1, patch_dim=(3,3)):
        ''' Take a new image and segment it '''

        # Shape of image
        W,H = I.shape

        # Compute minimum edges
        edgesW = (patch_dim[0]-1)/2
        edgesH = (patch_dim[1]-1)/2

        if clfr=='conv':
            # Take out patches
            Ip = aiu.images2patches(I.reshape((1,W,H)), patch_dim=patch_dim, edges=(edgesW,edgesH), stride=1)

            # Classify embedded patches
            Pp = np.argmax(model.predict(Ip),axis=1)

        elif clfr=='netw':
            # Take out patches
            Ip = self.propagate(self.image2allpatch(I,scan_ID=scan_ID,patch_dim=patch_dim))

            # Classify embedded patches
            Pp = model.predict(Ip)

        elif clfr=='flat':
            # Take out patches
            Ip = self.image2allpatch(I,scan_ID=scan_ID,patch_dim=patch_dim)[:,:-1]

            # Classify embedded patches
            Pp = model.predict(Ip)

        else:
            print('Unknown model')

        # Reassemble prediction image
        Pr = np.reshape(Pp, [W-(patch_dim[0]-1), H-(patch_dim[1]-1)])

        # Return re-assembled prediction
        return Pr



from distutils.core import setup
from setuptools import find_packages
from os.path import join, dirname


def read(fname):
    return open(join(dirname(__file__), fname)).read()


with open(join(dirname(__file__), 'mrainet/_version.py')) as versionpy:
    exec(versionpy.read())

with open('requirements.txt') as reqsfile:
    required = reqsfile.read().splitlines()

setup(
    name='mrainet',
    version=__version__,
    description=("MR acquisition-invariant network."),
    long_description=open('README.md').read(),
    packages=find_packages(),
    install_requires=required,
    url="https://github.com/wmkouw/mrai-net",
    license='Apache 2.0',
    author='Wouter Kouw',
    author_email='wmkouw@gmail.com',
    classifiers=['Topic :: Scientific/Engineering :: Artificial Intelligence',
                 'Topic :: Scientific/Engineering :: Image Recognition',
                 'License :: OSI Approved :: MIT License',
                 'Development Status :: 3 - Alpha',
                 'Operating System :: POSIX :: Linux',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6']
)



#!/usr/bin/env python3
import numpy as np
import numpy.random as rn
from scipy import sparse as sp

from keras import backend as kb
import keras.models as km
import keras.layers as kl
import keras.regularizers as kr
import sklearn.feature_extraction.image as sf


class MRAIConvolutionalNeuralNetwork(object):
    """
    Network for MRI-scanner acquisition-invariant representation learning.

    Class of convolutional neural networks that aim to map patches of two
    datasets from different MRI-scanners Methods include image processing
    operations, pair sampling and Siamese loss minimization.
    """

    def __init__(self, patch_size=(31, 31), classes=[1, 2, 3], num_draw=10,
                 num_kernels=[8], kernel_size=[(3, 3)], dense_size=[16, 8],
                 strides=(1, 1), dropout=0.1, l2=0.001, margin=1,
                 optimizer='rmsprop', batch_size=32, num_epochs=1):
        """
        Initialize with shape, architecture and optimization parameters.

        Parameters
        ----------
        patch_size: tuple(int, int)
            Dimensions of patches extracted from images
        classes: list[int]
            Numerical values of tissue classes
        num_draw : int
            Number of patches to draw from each pair of images
        num_kernels : list[int]
            Number of kernels in each layer, index corresponds to layer index
        kernel_size : list[(int, int)]
            Dimensions of convolution kernels in each layer
        dense_size : list[int]
            Width of the fully-connected layers
        strides : (int, int)
            The amount of pixels in each direction to skip while performing
            convolutions
        dropout : float
            Parameter controlling chance of randomly setting weights to 0.
            Dropout encourages convolution kernels to be independent of each
            other.
        l2 : float
            Parameter controlling the amount of regularization applied to the
            gradient update during training.
        margin : 1
            Parameter controlling how far apart dissimilar pairs should be.
        optimizer : str
            Type of optimizer to use for backpropagation.
        num_epochs : int
            Number of times the whole dataset is visited during stochastic
            gradient descent.

        Returns
        -------
        None

        """
        # Checks on data shapes
        if (patch_size[0] < 1) or (patch_size[1] < 1):
            raise ValueError('Negative patch size values.')
        if num_draw < 1:
            raise ValueError('Negative number of samples to draw.')

        # Store data shape parameters
        self.patch_size = patch_size
        self.num_draw = num_draw

        # Classes
        self.classes = classes

        # Check valid architecture sizes
        if (len(num_kernels) < 1) or (len(dense_size) < 1):
                raise ValueError('Too few layers specified.')

        # Network hyperparameters
        self.num_kernels = num_kernels
        self.kernel_size = kernel_size
        self.dense_size = dense_size
        self.strides = strides
        self.dropout = dropout
        self.margin = margin
        self.l2 = l2

        # Optimization parameters
        self.optimizer = optimizer
        self.batch_size = batch_size
        self.num_epochs = num_epochs

        # Initialize net architecture
        self.net = self.compile_net()

    def compile_net(self):
        """Compile network architecture."""
        # Setup input layers
        A = kl.Input(shape=(*self.patch_size, 1))
        B = kl.Input(shape=(*self.patch_size, 1))
        sIDA = kl.Input(shape=(1,))
        sIDB = kl.Input(shape=(1,))

        # Convolutional pipeline
        pipeC = km.Sequential()
        pipeC.add(kl.Conv2D(self.num_kernels[0],
                            self.kernel_size[0],
                            activation='relu',
                            padding='valid',
                            kernel_regularizer=kr.l2(self.l2),
                            input_shape=(*self.patch_size, 1)))
        pipeC.add(kl.MaxPooling2D(pool_size=(2, 2), padding='valid'))
        pipeC.add(kl.Dropout(self.dropout))
        for n in range(1, len(self.num_kernels)):
            pipeC.add(kl.Conv2D(self.num_kernels[n],
                                self.kernel_size[n],
                                activation='relu',
                                padding='valid',
                                kernel_regularizer=kr.l2(self.l2)))
            pipeC.add(kl.MaxPooling2D(pool_size=(2, 2), padding='valid'))
            pipeC.add(kl.Dropout(self.dropout))
        pipeC.add(kl.Flatten())

        # Dense pipeline
        pipeD = km.Sequential()
        pipeD.add(kl.Dense(self.dense_size[0],
                           activation='relu',
                           kernel_regularizer=kr.l2(self.l2),
                           input_shape=(pipeC.output_shape[1] + 1,)))
        for n in range(1, len(self.dense_size)):
            pipeD.add(kl.Dropout(self.dropout))
            pipeD.add(kl.Dense(self.dense_size[n],
                               activation='relu',
                               kernel_regularizer=kr.l2(self.l2)))
        pipeD.add(kl.Dense(2))

        # Distance in embedding space
        distance = kl.Lambda(self.l1_norm, output_shape=(1,))(
                          [pipeD(kl.concatenate([pipeC(A), sIDA])),
                           pipeD(kl.concatenate([pipeC(B), sIDB]))])

        # Set model
        model = km.Model(inputs=[A, B, sIDA, sIDB], outputs=distance)

        # Compile model with optimizer
        model.compile(loss=self.contrastive_loss, optimizer=self.optimizer)

        # Print model
        print(model.summary())

        # Return compiled model
        return model

    def contrastive_loss(self, label, distance):
        """
        Contrastive Siamese loss.

        For similar pairs, it consists of the squared Lp-distance.
        For dissimilar pairs, it consists of a hinge loss with respect to a
        margin parameter.

        Parameters
        ----------
        label : int
            Similarity label, 1=similar and 0=dissimilar
        distance: float
            Lp-norm between pairs of patches mapped through the network.

        Returns
        -------
        float
            Loss value for current pair of patches.

        """
        return kb.sum(label * kb.square(distance) +
                      (1-label) * kb.maximum(self.margin - distance, 0))

    def l1_norm(self, x):
        """l1-norm for loss layer."""
        return kb.sum(kb.abs(x[0] - x[1]), axis=1, keepdims=True)

    def l2_norm(self, x):
        """l2-norm for loss layer."""
        return kb.sum(kb.square(x[0] - x[1]), axis=1, keepdims=True)

    def gen_index_combs(self, x):
        """Generate combinations of two index arrays."""
        return np.array(np.meshgrid(x[0], x[1])).T.reshape(-1, 2)

    def matrix2sparse(self, X, edge=(0, 0), remove_nans=False):
        """
        Map matrix to a sparse array format.

        Parameters
        ----------
        X : array
            Matrix that should be mapped to sparse array format,
            may contain NaN's.
        edge : tuple(int, int)
            Dimensions of edge to ignore.
        remove_nans : bool
            Whether to remove NaN's as tissue labels.

        Returns
        -------
        sX : array
            Original matrix mapped to (i,j,v) format where i corresponds to the
            i-th row of X, j to the j-column of X and v of the value at
            position (i,j) of X.

        """
        # Shape
        h, w = X.shape

        # Remove edges from array
        X = X[edge[0]:h-edge[0], edge[1]:w-edge[1]]

        # Update shape
        h, w = X.shape

        # Generate meshgrid of coordinates
        tx, ty = np.meshgrid(np.arange(h), np.arange(w))

        # Reshape and stack index and value arrays
        sX = np.vstack((tx.T.ravel(), ty.T.ravel(), X.ravel())).T

        # Remove rows containing NaN's
        if remove_nans:
            sX = sX[~np.isnan(sX[:, 2]), :]

        return sX

    def subsample_rows(self, X, num_draw=1, replace=False):
        """
        Take a random subsample of rows from X.

        Parameters
        ----------
        X : array
            Array to subsample from.
        num_draw : int
            Number of rows to subsample.
        replace : bool
            Whether to replace sampled rows.

        Returns
        -------
        array
            Smaller array.

        """
        # Number of rows in array
        N = X.shape[0]

        # Check if number of samples exceeds size of array
        if (not replace) and (num_draw > N):
            raise ValueError('Number of samples larger than size array.')

        # Generate row range
        range_rows = np.arange(N)

        # Subsample from range
        index = rn.choice(range_rows, size=num_draw, replace=replace)

        # Return selected rows
        return X[index, :]

    def index2patch(self, X, index):
        """
        Slice patches from an image at given indices.

        Parameters
        ----------
        X : array
            input image
        index : array
            Row and column indices for the provided image.

        Returns
        -------
        patches : array
            Number of patches by patch height by patch width by 1.

        """
        # Check whether index is actually an array
        if not len(index.shape) > 0:
            raise ValueError('Index is an empty array.')

        # Number of patches to sample
        num_samples = index.shape[0]

        # Compute step length from patch center
        vstep = int((self.patch_size[0] - 1)/2)
        hstep = int((self.patch_size[1] - 1)/2)

        if len(index.shape) == 1:

            # Width and height indices
            w = np.zeros((num_samples,), dtype='uint8')
            h = np.zeros((num_samples,), dtype='uint8')

            # For every index, find its multilinear index
            for n in range(num_samples):
                w[n], h[n] = np.unravel_index(index[n], X.shape)

        elif len(index.shape) == 2:

            # First colum is row index, second column is is
            h = index[:, 0]
            w = index[:, 1]

        else:
            raise ValueError('Index has wrong shape.')

        # Preallocate patch array
        patches = np.zeros((num_samples, *self.patch_size, 1))

        # For each index, slice out subpatch
        for n in range(num_samples):
            patches[n, :, :, 0] = X[h[n] - vstep:h[n] + vstep + 1,
                                    w[n] - hstep:w[n] + hstep + 1]

        return patches

    def extract_random_patches(self, X, Y):
        """
        Extract a random set of patches from image.

        Parameters
        ----------
        X : array
            Input image to sample patches from.
        Y : array
            Label image corresponding to X.

        Returns
        -------
        patches : array
            Patches array, num patches by patch height by patch width by 1.
        labels : array
            Tissue label array corresponding to patches array.

        """
        # Shapes
        h, w = X.shape[1:3]
        num_images = X.shape[0]

        # Number of patches
        num_patches = num_images * len(self.classes) * self.num_draw

        # Compute step length from patch center
        vstep = int((self.patch_size[0] - 1)/2)
        hstep = int((self.patch_size[1] - 1)/2)

        # Preallocate input and output patches
        patches = np.zeros((num_patches, *self.patch_size, 1))
        labels = np.zeros((num_patches,), dtype='int64')

        # Define counter
        cnt = 0

        # Loop over number of images
        for i in range(num_images):

            # Loop over classes
            for k, classk in enumerate(self.classes):

                # Cut patch edges from image
                Yi = Y[i][vstep:h-vstep, hstep:w-hstep]

                # Find tissue in image
                Yi_k = np.argwhere(Yi == classk) + (vstep, hstep)

                # Subsample indices
                Yi_k = rn.choice(np.ravel_multi_index(Yi_k.T, (h, w)),
                                 size=self.num_draw, replace=False)

                # Current patch index
                ix = range(cnt*self.num_draw, (cnt+1)*self.num_draw)

                # Sample patch from class k
                patches[ix, :, :, :] = self.index2patch(X[i], Yi_k)

                # Store class labels
                labels[ix] = classk*np.ones((self.num_draw,))

                # Tick up
                cnt += 1

        return patches, labels

    def extract_all_patches(self, X, scan_ID=0, edge=(0, 0)):
        """
        Extract all patches from image, and append a scanner identification.

        Parameters
        ----------
        X : array
            Image to extract all patches from.
        scan_ID : int
            MRI-scanner identification variable.
        edge: tuple(int, int)
            Pixels past this edge should not be sampled from.

        Returns
        -------
        array
            Patches including scanner ID's, num_patches by patch size + 1.

        """
        # Shape
        h, w = X.shape

        # Remove edges from array
        X = X[edge[0]:h-edge[0], edge[1]:w-edge[1]]

        # Compute step length from patch center
        vstep = int((self.patch_size[0] - 1)/2)
        hstep = int((self.patch_size[1] - 1)/2)

        # Pad image before patch extraction
        X = np.pad(X, pad_width=((vstep, vstep), (hstep, hstep)),
                   mode='constant')

        # Sample patches at grid
        patches = sf.extract_patches_2d(X, patch_size=self.patch_size)

        # Number of patches extracted
        num_patches = patches.shape[0]

        # Augment patches with scanner ID and store
        return np.append(patches, scan_ID*np.ones((num_patches, 1)), axis=1)

    def sample_pairs(self, X, y, Z, u, num_draw=(10, 1)):
        """
        Sample a set of pairs of patches from two images.

        Parameters
        ----------
        X : array
            slice from source MRI-scanner
        y : array
            source tissue index sparse array; where each row i,j,k consists of
            the pixel's row index i, the pixel's column index j and the pixel's
            tissue k.
        Z : array
            slice from target MRI-scanner
        u : array
            target tissue index sparse array; where each row i,j,k consists of
            the pixel's row index i, the pixel's column index j and the pixel's
            tissue k.
        num_draw : tuple(int, int)
            maximum number of patches to draw from (source, target)

        Returns
        -------
        P : list[A, B, a, b]
            contains pairs of patches and scanner identifications
        S : array
            contains similarity labels between pairs

        """
        # Check if max draw does not exceed the number of supplied labels
        if (num_draw[0] > y.shape[0]) or (num_draw[1] > u.shape[0]):
            raise ValueError('Insufficient labels provided.')

        # Take only classes present in current slices
        classes = np.intersect1d(np.unique(y[:, 2]), np.unique(u[:, 2]))
        num_classes = len(classes)

        # Number of combinations per set of labels
        num_combs_yy = num_draw[0]**2
        num_combs_yu = num_draw[0]*num_draw[1]
        num_combs_uu = num_draw[1]**2

        # Total number of patches to draw
        total_num_draw = (num_combs_yy * num_classes*(num_classes - 1) *
                          num_combs_yu * num_classes*(num_classes - 1) *
                          num_combs_uu * num_classes*(num_classes - 1))

        # Preallocate patch arrays
        A = np.zeros((total_num_draw, *self.patch_size, 1))
        B = np.zeros((total_num_draw, *self.patch_size, 1))

        # Preallocate scanner label vectors
        a = np.zeros((total_num_draw, 1))
        b = np.zeros((total_num_draw, 1))

        # Preallocate similarity label array
        S = np.zeros((total_num_draw, 1))

        # Counter
        cnt = 0

        # Loop over classes
        for k, classk in enumerate(classes):

            # Take indices of current tissue
            index_yk = y[y[:, 2] == classk, :2].astype('uint8')
            index_uk = u[u[:, 2] == classk, :2].astype('uint8')

            # Subsample from index range
            index_yk = self.subsample_rows(index_yk, num_draw=num_draw[0])
            index_uk = self.subsample_rows(index_uk, num_draw=num_draw[1])

            # Map multilinear index to linear index
            index_yk = np.ravel_multi_index(index_yk.T, X.shape)
            index_uk = np.ravel_multi_index(index_uk.T, Z.shape)

            ''' Similar pairs from same source image '''

            # Generate pairwise index combinations
            combs = self.gen_index_combs((index_yk, index_yk))

            # Create current patch index pointer
            ix = slice(cnt*num_combs_yy, (cnt+1)*num_combs_yy)

            # Extract patches and store in arrays
            A[ix, :, :, :] = self.index2patch(X, combs[:, 0])
            B[ix, :, :, :] = self.index2patch(X, combs[:, 1])

            # Store scanner identifications of patches (X=0, Z=1)
            a[ix, :] = np.zeros((num_combs_yy, 1))
            b[ix, :] = np.zeros((num_combs_yy, 1))

            # Mark these combinations as similar(=1)
            S[ix, :] = np.ones((num_combs_yy, 1))

            # Increment counter
            cnt += 1

            ''' Similar pairs from source image and target image '''

            # Generate pairwise index combinations
            combs = self.gen_index_combs((index_yk, index_uk))

            # Create current patch index pointer
            ix = slice(cnt*num_combs_yu, (cnt+1)*num_combs_yu)

            # Extract patches and store in arrays
            A[ix, :, :, :] = self.index2patch(X, combs[:, 0])
            B[ix, :, :, :] = self.index2patch(Z, combs[:, 1])

            # Store scanner identifications of patches (X=0, Z=1)
            a[ix, :] = np.zeros((num_combs_yu, 1))
            b[ix, :] = np.ones((num_combs_yu, 1))

            # Mark these combinations as similar(=1)
            S[ix, :] = np.ones((num_combs_yu, 1))

            # Increment counter
            cnt += 1

            ''' Similar pairs from same target image '''

            # Generate pairwise index combinations
            combs = self.gen_index_combs((index_uk, index_uk))

            # Create current patch index pointer
            ix = slice(cnt*num_combs_uu, (cnt+1)*num_combs_uu)

            # Extract patches and store in arrays
            A[ix, :, :, :] = self.index2patch(Z, combs[:, 0])
            B[ix, :, :, :] = self.index2patch(Z, combs[:, 1])

            # Store scanner identifications of patches (X=0, Z=1)
            a[ix, :] = np.ones((num_combs_uu, 1))
            b[ix, :] = np.ones((num_combs_uu, 1))

            # Mark these combinations as similar(=1)
            S[ix, :] = np.ones((num_combs_uu, 1))

            # Increment counter
            cnt += 1

            # Loop over other classes
            for l, classl in enumerate(np.setdiff1d(classk, classes)):

                # Take indices of current tissue
                index_yl = y[y[:, 2] == classl, :1].astype('uint8')
                index_ul = u[u[:, 2] == classl, :1].astype('uint8')

                # Subsample indices
                index_yl = self.subsample_rows(index_yl, num_draw=num_draw[0])
                index_ul = self.subsample_rows(index_ul, num_draw=num_draw[1])

                # Map multilinear index to linear index
                index_yk = np.ravel_multi_index(index_yl.T, X.shape)
                index_uk = np.ravel_multi_index(index_ul.T, Z.shape)

                ''' Dissimilar pairs from same source image '''

                # Generate pairwise index combinations
                combs = self.gen_index_combs((index_yk, index_yl))

                # Create current patch index pointer
                ix = slice(cnt*num_combs_yy, (cnt+1)*num_combs_yy)

                # Extract patches and store in arrays
                A[ix, :, :, :] = self.index2patch(X, combs[:, 0])
                B[ix, :, :, :] = self.index2patch(X, combs[:, 1])

                # Store scanner identifications of patches (X=0, Z=1)
                a[ix, :] = np.zeros((num_combs_yy, 1))
                b[ix, :] = np.zeros((num_combs_yy, 1))

                # Mark these combinations as dissimilar(=0)
                S[ix, :] = np.zeros((num_combs_yy, 1))

                # Increment counter
                cnt += 1

                ''' Dissimilar pairs from source image and target image '''

                # Generate pairwise index combinations
                combs = self.gen_index_combs((index_yk, index_ul))

                # Create current patch index pointer
                ix = slice(cnt*num_combs_yu, (cnt+1)*num_combs_yu)

                # Extract patches and store in arrays
                A[ix, :, :, :] = self.index2patch(X, combs[:, 0])
                B[ix, :, :, :] = self.index2patch(Z, combs[:, 1])

                # Store scanner identifications of patches (X=0, Z=1)
                a[ix, :] = np.zeros((num_combs_yu, 1))
                b[ix, :] = np.ones((num_combs_yu, 1))

                # Mark these combinations as dissimilar(=0)
                S[ix, :] = np.zeros((num_combs_yu, 1))

                # Increment counter
                cnt += 1

                ''' Dissimilar pairs from same target image '''

                # Generate pairwise index combinations
                combs = self.gen_index_combs((index_uk, index_ul))

                # Create current patch index pointer
                ix = slice(cnt*num_combs_uu, (cnt+1)*num_combs_uu)

                # Extract patches and store in arrays
                A[ix, :, :, :] = self.index2patch(Z, combs[:, 0])
                B[ix, :, :, :] = self.index2patch(Z, combs[:, 1])

                # Store scanner identifications of patches (X=0, Z=1)
                a[ix, :] = np.ones((num_combs_uu, 1))
                b[ix, :] = np.ones((num_combs_uu, 1))

                # Mark these combinations as dissimilar(=0)
                S[ix, :] = np.zeros((num_combs_uu, 1))

                # Increment counter
                cnt += 1

        # Return collected patches
        return [A, B, a, b], S

    def train(self, X, Y, Z, U, num_targets=1):
        """
        Train the network using pairs of patches from the images.

        Parameters
        ----------
        X : array
            source scans, slices by height by width
        Y : array
            source labels, slices by height by width
        Z : array
            target scans, slices by height by width
        U : array
            target labels, slices by height by width,
            contains NaN's at unknown labels
        num_targets : int
            How many target labels to use.

        Returns
        -------
        None

        """
        # Number of subjects from each scanner
        num_src_sub = X.shape[0]
        num_tgt_sub = Z.shape[0]

        # Loop over source images
        for i in range(num_src_sub):
            print('At source subject '+str(i+1)+'/'+str(num_src_sub))

            # Map source label image to sparse array
            Yi = self.matrix2sparse(Y[i])

            # Loop over target images
            for j in range(num_tgt_sub):

                # Map target label image to sparse array
                Uj = self.matrix2sparse(U[j])

                # Number of target labels of current image
                num_labels = Uj.shape[0]

                # Check number of targets does not exceed number of labels
                if num_targets > num_labels:
                    raise ValueError('More targets than labels asked.')

                # Draw pairs of patches from two images
                P, S = self.sample_pairs(X[i], Yi, Z[j], Uj,
                                         num_draw=(self.num_draw, num_targets))

                # Train on current set
                self.net.fit(x=P, y=S,
                             epochs=self.num_epochs,
                             batch_size=self.batch_size,
                             validation_split=0.1,
                             shuffle=True,
                             verbose=1)

            # Check for multiple source subjects
            if num_src_sub > 1:

                # Take random other source image
                o = np.random.choice(np.setdiff1d(np.arange(num_src_sub), i),
                                     size=1, replace=False)[0]

                # Map target label image to sparse array
                Yo = self.matrix2sparse(Y[o])

                # Draw pairs of patches from two images
                P, S = self.sample_pairs(X[i], Yi, X[o], Yo,
                                         num_draw=(self.num_draw, 
                                                   self.num_draw))

                # Train on current set
                self.net.fit(x=P, y=S,
                             epochs=self.num_epochs,
                             batch_size=self.batch_size,
                             validation_split=0.1,
                             shuffle=True,
                             verbose=1)

        # Report
        print('Training complete.')

    def save_model(self, model_fn, weights_fn):
        """
        Save model to filename.

        Parameters
        ----------
        model_fn : str
            Filename to save model to.
        weights_fn : str
            Filename to save weight matrix to.

        Returns
        -------
        None

        """
        # Check if model exists
        if self.net:

            # Serialize model to JSON
            model_json = self.net.to_json()

            # Write to file
            with open(model_fn, 'w') as json_file:
                json_file.write(model_json)

            # Serialize weights to HDF5
            self.net.save_weights(weights_fn)

            # Report
            print('Saved model to disk')

        else:
            raise ValueError('No model compiled yet.')

    def load_model(self, model_fn, weights_fn):
        """
        Load model from filename.

        Parameters
        ----------
        model_fn : str
            Filename of saved model.
        weights_fn : str
            Filename of saved weight matrix.

        Returns
        -------
        None

        """
        # Read json model file
        with open(model_fn, 'r') as json_file:
            loaded_model_json = json_file.read()

        # Construct model from json file
        self.net = km.model_from_json(loaded_model_json)

        # Load weights into new model
        self.net.load_weights(weights_fn)

        # Report
        print("Loaded model from disk")

    def feedforward(self, patches, scan_ID):
        """
        Feed a set of patches forward through the network.

        Parameters
        ----------
        patches : array
            Contains patches in form of number of patches by patch height
            by patch width by 1.
        scan_ID : int
            Scanner identification variable, indicating from which MRI-scanner
            these patches came from.

        Returns
        -------
        array
            Final layer representation of patches fed forward through the
            network.

        """
        # Check whether patches have number of channels
        if len(patches.shape) < 4:
            raise ValueError('Patches array has less than 4 dimensions.')

        # Number of patches
        num_patches = patches.shape[0]

        # Define intermediate-output model
        interm_model = km.Model(inputs=self.net.input,
                                outputs=self.net.layers[-2].get_output_at(1))

        # Scan ID list
        sID = scan_ID*np.ones((num_patches, 1))

        # Feed forward
        return interm_model.predict([patches, patches, sID, sID])

    def segment_image(self, X, model, feed=True, mapost=False, scan_ID=1):
        """
        Segment a new image using the trained network.

        Parameters
        ----------
        X : array
            new image that needs to be segmented.
        model : sklearn-model
            Trained classifier from scikit-learn,
            needs to have a predict method.
        feed : bool
            whether the extracted patches should be fed through the network,
            a value of False is for experimental purposes (def: True)
        mapost : bool
            whether to map the predictions to a maximum a posteriori form.
            (def: False)
        scan_ID : int
            scanner identification of new image.

        Returns
        -------
        preds : array
            Label image of same size as input image, containing predictions
            made by the provided trained classifier.

        """
        # Shape of image
        imsize = X.shape

        # Take out patches
        patches = self.extract_all_patches(X, scan_ID=scan_ID)

        # Feed through network
        if feed:
            patches = self.feedforward(patches, scan_ID=scan_ID)

        # Classify embedded patches
        preds = model.predict(patches)

        # If model outputs posteriors, then take maximum a posteriori
        if mapost:
            preds = np.argmax(preds, axis=1) + np.min(self.classes)

        # Reassemble prediction image
        return np.reshape(preds, imsize)
